from typing import TypeVar, Union
from unittest.mock import _Call  # noqa


class MISSING_TYPE:
    """
    Class to indicate a missing value
    """


_T = TypeVar("_T")


Opt = Union[_T, MISSING_TYPE]

MISSING = MISSING_TYPE()

Call = _Call
call = Call(from_kall=False)

__all__ = ["MISSING", "call"]



ADJECTIVES = [
    "adorable",
    "adventurous",
    "affectionate",
    "ambitious",
    "amusing",
    "anxious",
    "artistic",
    "attentive",
    "beautiful",
    "bold",
    "brave",
    "bright",
    "brilliant",
    "calm",
    "careful",
    "cautious",
    "charming",
    "cheerful",
    "clever",
    "compassionate",
    "confident",
    "considerate",
    "cooperative",
    "courageous",
    "courteous",
    "creative",
    "curious",
    "daring",
    "delightful",
    "determined",
    "diligent",
    "eager",
    "efficient",
    "empathetic",
    "energetic",
    "enthusiastic",
    "friendly",
    "funny",
    "generous",
    "gentle",
    "graceful",
    "grateful",
    "hardworking",
    "helpful",
    "honest",
    "hopeful",
    "humble",
    "imaginative",
    "independent",
    "innovative",
    "insightful",
    "intelligent",
    "intuitive",
    "inventive",
    "jovial",
    "joyful",
    "kind",
    "knowledgeable",
    "loving",
    "loyal",
    "meticulous",
    "modest",
    "motivated",
    "observant",
    "optimistic",
    "organized",
    "passionate",
    "patient",
    "peaceful",
    "perceptive",
    "persistent",
    "playful",
    "polite",
    "practical",
    "proactive",
    "punctual",
    "quick-witted",
    "reliable",
    "resilient",
    "respectful",
    "romantic",
    "sensitive",
    "sincere",
    "skillful",
    "sociable",
    "spirited",
    "spontaneous",
    "supportive",
    "sympathetic",
    "thoughtful",
    "tidy",
    "trustworthy",
    "understanding",
    "unique",
    "versatile",
    "warm",
    "witty",
    "wise",
    "zany",
    "zealous",
]

NOUNS = [
    "apple",
    "backpack",
    "balloon",
    "banana",
    "beach",
    "bear",
    "bed",
    "bicycle",
    "bird",
    "blanket",
    "boat",
    "book",
    "bottle",
    "bridge",
    "butterfly",
    "camera",
    "candle",
    "car",
    "castle",
    "cat",
    "chair",
    "cheese",
    "city",
    "cloud",
    "computer",
    "couch",
    "cow",
    "cup",
    "deer",
    "dog",
    "dolphin",
    "door",
    "dragon",
    "dream",
    "dress",
    "elephant",
    "factory",
    "fire",
    "fish",
    "flower",
    "forest",
    "fox",
    "frog",
    "garden",
    "giraffe",
    "glasses",
    "guitar",
    "hat",
    "horse",
    "house",
    "island",
    "jacket",
    "jellyfish",
    "kangaroo",
    "key",
    "kite",
    "lamp",
    "leaf",
    "lion",
    "lizard",
    "mountain",
    "mouse",
    "mug",
    "ocean",
    "owl",
    "panda",
    "parrot",
    "pen",
    "pencil",
    "phone",
    "piano",
    "pillow",
    "planet",
    "plate",
    "pocket",
    "rabbit",
    "rainbow",
    "river",
    "robot",
    "rocket",
    "rose",
    "sandwich",
    "scarf",
    "shoe",
    "skateboard",
    "snake",
    "snowflake",
    "sock",
    "spoon",
    "star",
    "sun",
    "sunglasses",
    "table",
    "teapot",
    "teddy",
    "tiger",
    "tree",
    "turtle",
    "umbrella",
    "violin",
    "watch",
    "watermelon",
    "whale",
    "window",
    "zebra",
]



from typing import Any
import pytest

from megamock.megapatches import MegaPatch


def pytest_sessionstart(*args, **kwargs) -> None:
    import megamock

    megamock.start_import_mod()


def pytest_addoption(parser: Any) -> None:
    parser.addoption(
        "--do_not_autostop_megapatches",
        help="Disable autostopping MegaPatches after every test",
        action="store",
        default=False,
    )


@pytest.fixture(autouse=True)
def stop_all_megapatches(request) -> None:
    if not request.config.getoption("--do_not_autostop_megapatches"):
        MegaPatch.stop_all()


# swap out default mocker if pytest-mock is installed
try:
    from pytest_mock import MockerFixture  # type: ignore  # noqa  # test for install

    @pytest.fixture(autouse=True, scope="session")
    def use_pytest_mocker(session_mocker: MockerFixture) -> None:
        MegaPatch.default_mocker = session_mocker

except ImportError:
    pass  # not installed



from __future__ import annotations

import functools
import inspect
import logging
import sys
from functools import cached_property
from types import ModuleType
from typing import Any, Callable, Generic, Iterable, TypeVar, cast
from unittest import mock

from varname import argname  # type: ignore

from megamock.import_references import References
from megamock.megamocks import MegaMock, _MegaMockMixin, _UseRealLogic

logger = logging.getLogger(__name__)

T = TypeVar("T")
U = TypeVar("U")


class _MISSING:
    """
    Class to indicate a missing argument
    """


class MegaPatchBehavior:
    def __init__(
        self,
        *,
        autospec: bool,
    ) -> None:
        """
        Define the mocking behavior.

        By default, MegaMock.it(...) will determine the behavior from the passed in
        type. Pass in this class as the `behavior` argument to explicitly set the
        behavior.

        :param autospec: Autospec the thing being mocked
        """
        self.autospec = autospec

    @staticmethod
    def for_thing(thing: Any) -> MegaPatchBehavior:
        return MegaPatchBehavior(autospec=True)


class MegaPatch(Generic[T, U]):
    __reserved_names = {"_patches", "_thing", "_path", "_mocked_value", "_return_value"}
    _active_patches: set[MegaPatch] = set()

    default_mocker: ModuleType | object = mock

    def __init__(
        self,
        *,
        thing: Any,
        patches: list[mock._patch],
        new_value: MegaMock | Any,
        return_value: Any,
        mocker: ModuleType | object
        # _merged_type: type[U] | None = None,
    ) -> None:
        self._patches = patches
        self._thing: Any | None = thing
        self._new_value: MegaMock = new_value
        self._return_value = return_value
        self._mocker = mocker

        self._started = False

    @property
    def patches(self) -> list[mock._patch]:
        return list(self._patches)

    @property
    def thing(self) -> Any:
        return self._thing

    @property
    def new_value(self) -> MegaMock[T, U] | Any:
        return self._new_value

    def new_value_is_mock(self) -> bool:
        val = self.new_value
        return isinstance(val, MegaMock) or hasattr(val, "return_value")

    @property
    def mock(self) -> MegaMock[T, U]:
        if not self.new_value_is_mock():
            raise ValueError(f"New value {self.new_value!r} is not a mock!")
        return self.new_value

    @property
    def megainstance(self) -> U:
        return self.mock.megainstance

    @property
    def return_value(self) -> MegaMock[T, U]:
        return cast(MegaMock[T, U], self._return_value)

    @return_value.setter
    def return_value(self, new_value: MegaMock[T, U]) -> None:
        self._return_value = new_value
        self._new_value.return_value = new_value

    def set_context_manager_return_value(self, new_return_value: Any) -> None:
        """
        Use to modify the result of entering a context manager

        megapatch = MegaPatch.it(my_context_manager)
        megapatch.set_context_manager_return_value("foo")

        with my_context_manager() as val:
            assert val == "foo"
        """
        try:
            self.return_value.__enter__.return_value = new_return_value  # type: ignore
        except AttributeError:
            raise ValueError("Not a context manager")

    def set_context_manager_side_effect(self, new_side_effect: Any) -> None:
        """
        Use to set the side effect when entering the context manager. As with
        normal side-effect usage, an exception will result in it being raised,
        and an iterable will change the result on subsequent calls
        """
        try:
            self.return_value.__enter__.side_effect = new_side_effect  # type: ignore
        except AttributeError:
            raise ValueError("Not a context manager")

    def set_context_manager_exit_side_effect(self, new_side_effect: Any) -> None:
        """
        Use to set the side effect when exiting the context manager. As with
        normal side-effect usage, an exception will result in it being raised,
        and an iterable will change the result on subsequent calls
        """
        try:
            self.return_value.__exit__.side_effect = new_side_effect  # type: ignore
        except AttributeError:
            raise ValueError("Not a context manager")

    def start(self) -> None:
        if self._started:
            return
        # support for pytest-mock and similar
        if not hasattr(self._mocker, "stopall"):
            # built-in mock
            for patch in self._patches:
                patch.start()
        MegaPatch._active_patches.add(self)
        self._started = True

    def stop(self) -> None:
        # support for pytest-mock and similar
        if hasattr(self._mocker, "stopall"):
            self._mocker.stopall()
        else:
            # built-in mock
            for patch in self._patches:
                patch.stop()
        MegaPatch._active_patches.remove(self)
        self._started = False

    def __enter__(self) -> MegaPatch[T, U]:
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        self.stop()

    def __call__(self, func) -> Callable[..., Any]:
        """
        Decorator to patch a function
        """

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            self.start()
            try:
                return func(*args, **kwargs)
            finally:
                self.stop()

        return wrapper

    @staticmethod
    def active_patches() -> list[MegaPatch]:
        return list(MegaPatch._active_patches)

    @staticmethod
    def stop_all() -> None:
        for megapatch in list(MegaPatch._active_patches):
            megapatch.stop()

    @staticmethod
    def _get_new_and_return_value_with_autospec(
        behavior: MegaPatchBehavior,
        thing: Any,
        spec_set: bool,
        return_value: Any,
        side_effect: Iterable | Exception | None,
    ) -> tuple[Any, Any]:
        if behavior.autospec:
            autospeced = mock.create_autospec(thing, spec_set=spec_set)
            if inspect.isfunction(autospeced):
                assert hasattr(autospeced, "return_value")
                if return_value is not _MISSING:
                    autospeced.return_value = return_value
                if side_effect is not None:
                    autospeced.side_effect = side_effect  # type: ignore
                new = autospeced
            else:
                new = MegaMock.from_legacy_mock(autospeced, spec=thing)
            return_value = new.return_value
        else:
            if return_value is _MISSING:
                return_value = MegaMock()
            new = MegaMock(return_value=return_value)
        return new, return_value

    @staticmethod
    def it(
        thing: T,
        /,
        new: Any | None = None,
        spec_set: bool = True,
        behavior: MegaPatchBehavior | None = None,
        autostart: bool = True,
        mocker: ModuleType | object | None = None,
        new_callable: Callable | None = None,
        **kwargs: Any,
    ):
        """
        MegaPatch something.

        :param thing: The thing to patch
        :param spec_set: If true, then raise an attribute error when setting an
            attribute that doesn't exist on a class
        :param behavior: The behavior to use when mocking
        :param autostart: If true, then start the patch immediately
        :param mocker: The object to use for patching. If None, then use the default
        :param new_callable: A callable function or object that returns the replacement
            object to return. This is usually some replacement Mock bject.
            This is mainly for legacy support and is not recommended since it can't be
            combined with autospec.
        """
        if mocker is None:
            mocker = MegaPatch.default_mocker
        else:
            assert hasattr(
                mocker, "patch"
            ), "mocker does not appear to be a Mocker object"

        if autostart is False and not hasattr(mocker.patch, "start"):
            logger.warning(
                "Disabling autostart doesn't appear to be supported by mocker. "
                "Falling back to built in mock"
            )
            mocker = mock

        if isinstance(thing, _MegaMockMixin):
            parent_mock = thing.megamock.parent
            assert thing.megamock.spec is not None
            thing = thing.megamock.spec
        else:
            parent_mock = None

        if behavior is None:
            behavior = MegaPatchBehavior.for_thing(thing)
        if new_callable is not None:
            if new is None:
                new = mock.DEFAULT
            return_value = kwargs.get("return_value")
            kwargs["new_callable"] = new_callable
        else:
            if (autospec := kwargs.pop("autospec", None)) in (True, False):
                behavior.autospec = autospec
            new, return_value = MegaPatch._new_return_value(
                thing, spec_set, new, kwargs, behavior
            )
        if isinstance(thing, cached_property):
            thing = thing.func  # type: ignore

        passed_in_name = argname("thing", func=MegaPatch.it, vars_only=False)
        corrected_passed_in_name = MegaPatch._correct_for_renamed_import(
            passed_in_name, thing
        )

        name_to_patch, module_path = MegaPatch._determine_module_path_and_name(
            thing, passed_in_name, corrected_passed_in_name
        )

        patches = MegaPatch._build_patches(
            mocker, module_path, name_to_patch, corrected_passed_in_name, new, kwargs
        )

        mega_patch = MegaPatch[T, type[MegaMock | T]](
            thing=thing,
            patches=patches,
            new_value=new,
            return_value=return_value,
            mocker=mocker,
        )
        if autostart:
            mega_patch.start()

        MegaPatch._maybe_assign_link(parent_mock, corrected_passed_in_name, mega_patch)

        return mega_patch

    @staticmethod
    def _correct_for_renamed_import(passed_in_name: str, thing: Any) -> str:
        qualname = getattr(thing, "__qualname__", None)
        if qualname is None:
            module_name = MegaPatch._get_module_path_for_nonclass()
            return References.get_original_name(module_name, passed_in_name)
        return qualname

    @staticmethod
    def _maybe_assign_link(
        parent_mock: _MegaMockMixin | None, passed_in_name: str, mega_patch: MegaPatch
    ) -> None:
        if parent_mock is not None:
            assert passed_in_name
            this_name = passed_in_name.split(".")[-1]
            try:
                cast(
                    MegaMock, getattr(parent_mock.megainstance, this_name)
                )._megalink_to(mega_patch.mock)
            except ValueError:
                pass  # not a mock

    @staticmethod
    def _new_return_value(
        thing: Any,
        spec_set: bool,
        new_given: Any,
        kwargs: dict,
        behavior: MegaPatchBehavior,
    ) -> tuple[Any, Any]:
        provided_return_value = kwargs.pop("return_value", _MISSING)
        side_effect = kwargs.pop("side_effect", None)

        MegaPatch._gotcha_check(provided_return_value, behavior)

        if new_given is None:
            if behavior.autospec:
                new, return_value = MegaPatch._get_new_and_return_value_with_autospec(
                    behavior,
                    thing,
                    spec_set,
                    provided_return_value,
                    side_effect=side_effect,
                )
            else:
                if provided_return_value is _MISSING:
                    return_value = MegaMock()
                else:
                    return_value = provided_return_value
                new = MegaMock[None, None](return_value=return_value)
        else:
            new = new_given
            if provided_return_value is not _MISSING:
                logger.warning("Ignoring return_value argument when 'new' is provided")
            if hasattr(new, "return_value"):
                return_value = new.return_value
            else:
                return_value = None

        assert return_value is not _MISSING

        return new, return_value

    @staticmethod
    def _gotcha_check(return_value: Any, behavior: MegaPatchBehavior) -> None:
        # autospec does not use the MegaMock code path when building return values
        # so this does not work as expected. It ends up returning a UseRealLogic object
        if behavior.autospec and isinstance(return_value, _UseRealLogic):
            raise ValueError(
                "Setting the return value within MegaPatch.it "
                "is currently not supported"
            )

    @staticmethod
    def _determine_module_path_and_name(
        thing: Any, passed_in_name: str, corrected_passed_in_name: str
    ) -> tuple[str, str]:
        if not (module_path := getattr(thing, "__module__", None)):
            owning_class = MegaPatch._get_owning_class(passed_in_name)
            if owning_class:
                return corrected_passed_in_name, owning_class.__module__
        if module_path is None:
            module_path = MegaPatch._get_module_path_for_nonclass()
            if module_path is None:
                raise Exception(f"Unable to determine module path for: {thing!r}")
            return passed_in_name, module_path
        return corrected_passed_in_name, module_path

    @staticmethod
    def _build_patches(
        mocker: Any,
        module_path: str,
        name_to_patch: str,
        corrected_passed_in_name: str,
        new: Any,
        kwargs: dict,
    ) -> list[mock._patch]:  # type: ignore  # mypy bug?
        patches = []
        for path, named_as in (
            References.get_references(module_path, corrected_passed_in_name)
            | References.get_reverse_references(module_path, corrected_passed_in_name)
            | {(module_path, name_to_patch)}
        ):
            mock_path = f"{path}.{named_as}"
            p = mocker.patch(mock_path, new, **kwargs)
            patches.append(p)

        return patches

    @staticmethod
    def _get_module_path_for_nonclass() -> str:
        stack = inspect.stack()
        module = inspect.getmodule(stack[3][0])
        assert module
        return module.__name__

    @staticmethod
    def _get_owning_class(name: str) -> str | None:
        if "." not in name:
            return None
        owning_class_name, attr_name = name.rsplit(".", 1)
        calling_frame = sys._getframe(1)
        return calling_frame.f_locals.get(owning_class_name, None)



from .import_machinery import start_import_mod
from .megamocks import MegaMock
from .megapatches import MegaPatch
from .megas import Mega
from .type_util import Call

__all__ = [
    "Call",
    "Mega",
    "MegaMock",
    "MegaPatch",
    "start_import_mod",
    "UseRealLogic",
]



from __future__ import annotations

import copy
import random
import re
import time
import traceback
from abc import ABCMeta
from collections import defaultdict
from dataclasses import dataclass, field
from inspect import isawaitable, isclass, iscoroutinefunction
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    TypeVar,
    cast,
    get_type_hints,
    overload,
)
from unittest import mock
from unittest.mock import create_autospec

from megamock import name_words
from megamock.type_util import MISSING, MISSING_TYPE

T = TypeVar("T")
U = TypeVar("U")


class SpecRequiredException(Exception):
    """
    Raised when a spec is required but not provided
    """

    def __str__(self) -> str:
        return "Spec is required but not provided"


class _UseRealLogic:
    """
    Class to indicate that the real logic should be used
    """


UseRealLogic = _UseRealLogic()


class AttributeTrackingBase(metaclass=ABCMeta):
    stacktrace: list[traceback.FrameSummary]

    @property
    def top_of_stacktrace(self) -> list[str]:
        """
        Convenience property for quickly viewing the stacktrace in an IDE debugger.

        This displays the folder / file and then the rest of the stacktrace
        """
        ret = []
        for x in self.format_stacktrace(5):
            if rslt := re.search(
                r"[^/\\]+[/\\][\w\.]+\", line \d+,.*", x, re.MULTILINE | re.DOTALL
            ):
                lines = rslt.group(0).splitlines()
                lines[0] = "..." + lines[0]
                ret.extend([line for line in lines if line])
            else:
                ret.append(x)
        return ret

    def format_stacktrace(self, max_depth=100) -> list[str]:
        # for lists, the first frame is the most recent
        return traceback.format_list(self.stacktrace[:max_depth])

    def print_stacktrace(self, max_depth=100) -> None:
        # when printing stacktraces, display the most recent frame last
        traceback.print_list(self.stacktrace[:-max_depth:-1])


class AttributeAssignment(AttributeTrackingBase):
    def __init__(
        self, attr_name: str, attr_value: Any, stacktrace: list[traceback.FrameSummary]
    ) -> None:
        self.attr_name = attr_name
        self.attr_value = attr_value
        self.stacktrace = stacktrace  # 0 is most recent frame, not oldest frame
        self.time = time.time()

    @staticmethod
    def for_current_stack(
        attr_name: str, attr_value: Any, starting_depth=3
    ) -> AttributeAssignment:
        stacktrace = traceback.extract_stack()[-starting_depth::-1]
        return AttributeAssignment(attr_name, attr_value, stacktrace)


class SpyAccess(AttributeTrackingBase):
    def __init__(
        self, attr_name: str, attr_value: Any, stacktrace: list[traceback.FrameSummary]
    ) -> None:
        self.attr_name = attr_name
        self.attr_value = copy.copy(
            attr_value  # just doing shallow copy, may need to revisit in future
        )
        self.stacktrace = stacktrace  # 0 is most recent frame, not oldest frame
        self.time = time.time()

    @staticmethod
    def for_current_stack(
        attr_name: str, attr_value: Any, starting_depth=3
    ) -> SpyAccess:
        stacktrace = traceback.extract_stack()[-starting_depth::-1]
        return SpyAccess(attr_name, attr_value, stacktrace)


_base_mock_types = (
    mock.Mock
    | mock.MagicMock
    | mock.NonCallableMock
    | mock.NonCallableMagicMock
    | mock.AsyncMock
)


@dataclass
class MegaMockAttributes:
    parent: _MegaMockMixin | None = None
    spec: Any | None = None
    wraps: Any | None = None
    spy: Any | None = None
    attr_assignments: dict[str, list[AttributeAssignment]] = field(
        default_factory=lambda: defaultdict(list)
    )
    spied_access: dict[str, list[SpyAccess]] = field(
        default_factory=lambda: defaultdict(list)
    )
    name: str | None = None

    _wrapped_mock: _base_mock_types | None = None


class _MegaMockMixin(Generic[T, U]):
    """
    Mixin used by MegaMock, NonCallableMegaMock, and AsyncMegaMock
    """

    USE_SUPER_SETATTR = {
        # properties where the actual value uses a different name
        # __setattr__ comes before property evaluation
        "side_effect",
        "called",
        "call_count",
        "call_args",
        "call_args_list",
        "await_count",
        "await_args",
        "await_args_list",
        "return_value",
    }

    meganame: str  # a human friendly name, "<ADJECTIVE> <VERB> <INTEGER>"
    megamock: MegaMockAttributes = None  # type: ignore
    _wrapped_legacy_mock: _base_mock_types | None
    _mock_return_value_cache: Any | MISSING_TYPE = None
    # link call behavior to this mock
    _linked_mock: MegaMock[T, U] | None = None

    def __init__(
        self,
        spec: T | type[T] | None = None,
        *,
        wraps: T | None = None,
        spy: T | None = None,
        spec_set: bool = True,
        instance: bool | None = None,
        side_effect: Any = None,
        return_value: Any = MISSING,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ) = None,
        _parent_mega_mock: _MegaMockMixin | None = None,
        _name: str | None = None,
        _merged_type: type[U] | None = None,
        # warning: kwargs to MagicMock may not work correctly! Use at your own risk!
        **kwargs,
    ) -> None:
        """
        :param spec: The MegaMock's specification (template object)
        :param wraps: An object to wrap, this is included for legacy support. Prefer spy
        :param spy: An object to spy on. The mock will maintain the real behavior of the
            object, but will also track attribute access and assignments
        :param spec_set: If True, only attributes in the spec will be allowed. Assigning
            attributes not part of the spec will result in a AttributeError
        :param instance: If True, the mock will be an instance of the spec. If False,
            the mock will be a class. By default this is True. This must be omitted or
            False for the AsyncMock
        :param side_effect: The side effect to use for the mock. Exceptoins are raised,
            fuctions are called, and iterables are returned in order in subsequent calls
        :param return_value: The return value to use for the mock.
        :param _wraps_mock: the wrapped mock, for internal use
        :param _parent_mega_mock: The parent MegaMock, for internal use
        :param _name: The name of the mock, for internal use
        """
        self.meganame = self._generate_meganame()

        self._linked_mock = None
        megamock_attrs = MegaMockAttributes()
        megamock_attrs.name = self._generate_mock_name(spec, _parent_mega_mock, _name)
        self._wrapped_legacy_mock = None
        self._mock_return_value_cache = MISSING

        megamock_attrs.parent = _parent_mega_mock
        if instance is None:
            if iscoroutinefunction(spec) or isawaitable(spec):
                instance = False
            else:
                instance = True

        if wraps and spy:
            # if spy is used, then the spied value is also wrapped
            raise Exception("Cannot both wrap and spy")

        megamock_attrs.wraps = wraps = wraps or spy
        megamock_attrs.spec = spec
        if wraps and not _wraps_mock:
            _wraps_mock = mock.MagicMock(wraps=wraps)

        megamock_attrs.spy = spy
        if return_value is not MISSING:
            kwargs["return_value"] = return_value
        kwargs["side_effect"] = side_effect

        if _wraps_mock is None:
            if spec is not None:
                autospeced_legacy_mock = mock.create_autospec(
                    spec, spec_set=spec_set, instance=instance, **kwargs
                )
                megamock_attrs._wrapped_mock = autospeced_legacy_mock
            else:
                # not wrapping a Mock object, so do init for super
                # this is not done when __wrapped is set because things like
                # mock calls occur on the wrapped object, but calling init
                # would set values like call_args_list on this object
                super().__init__(**kwargs)
        else:
            # there is a bug where simply checking for "return_value"
            # on a MagicMock will set it, breaking the wrapped value
            if not wraps:
                self._mock_return_value_ = _wraps_mock.return_value
            else:
                self._mock_return_value_ = None
            megamock_attrs._wrapped_mock = _wraps_mock

        # must be last as the setattr behavior will change after this
        self.megamock = megamock_attrs
        # shortcut to the wrapped mock to avoid performance penalty
        self._wrapped_legacy_mock = megamock_attrs._wrapped_mock

    def _generate_mock_name(
        self,
        spec: Any,
        parent_mega_mock: _MegaMockMixin | None,
        name: str | None,
    ) -> str:
        mock_name = ""
        if spec is None:
            mock_name = "MegaMock()"
        if parent_mega_mock is not None:
            mock_name = parent_mega_mock.megamock.name or "MegaMock()"
            mock_name += f".{name}" if name else "()"
        try:
            if spec is not None:
                spec_name = getattr(spec, "__name__", None)
                if spec_name is not None:
                    if mock_name:
                        mock_name += "."
                    mock_name += spec_name
                else:
                    if mock_name:
                        mock_name += " -> "
                    mock_name += spec.__class__.__name__
        except AttributeError:
            mock_name += "mock"
            if parent_mega_mock:
                mock_name += "()"
        return mock_name

    def _generate_meganame(self) -> str:
        """
        Generate a random adjective / noun / number name for the mock
        """
        return (
            f"{random.choice(name_words.ADJECTIVES)} "
            f"{random.choice(name_words.NOUNS)} "
            f"{random.randint(0, 10000)}"
        )

    @property
    def megainstance(self) -> U:
        """
        Access the instance of a class mock.
        """
        if not isinstance(
            self.megamock._wrapped_mock,
            (mock.MagicMock, mock.Mock),  # callable mocks
        ) or not isclass(self.megamock.spec):
            raise Exception("The megainstance property was intended for class mocks")
        return cast(U, self.return_value)

    @property
    def _mock_return_value(self) -> Any:
        if self._wrapped_legacy_mock is not None:
            if (val := self._mock_return_value_cache) is MISSING:
                if isinstance(
                    self._mock_return_value_,
                    mock.NonCallableMagicMock | mock.NonCallableMock,
                ):
                    val = self._mock_return_value_cache = MegaMock.from_legacy_mock(
                        self._mock_return_value_,
                        None,
                        self.megamock.wraps,
                    )
                else:
                    val = self._mock_return_value_cache = self._mock_return_value_
            return val
        return self.__dict__.get(
            "_mock_return_value", self.__class__._mock_return_value
        )

    def _get_child_mock(self, /, **kw) -> MegaMock:
        return MegaMock(_parent_mega_mock=self, **kw)

    def __getattr__(self, key) -> Any:
        if key != "megamock" and self.megamock.spy is not None:
            result = getattr(self.megamock.spy, key)
            # if result is callable let wrapped handle it so that it's a mock object
            if not callable(result):
                self.megamock.spied_access[key].append(
                    SpyAccess.for_current_stack(key, result)
                )
                return result

        if key == "_spec_signature":
            return self.__signature__

        if (wrapped := self._wrapped_legacy_mock) is not None:
            result = getattr(wrapped, key)
            if not isinstance(result, _MegaMockMixin) and isinstance(
                result, mock.NonCallableMock | mock.NonCallableMagicMock
            ):
                mega_result = MegaMock.from_legacy_mock(
                    result,
                    getattr(self.megamock.spec, key, self.megamock.spec),
                    self.megamock.wraps,
                    parent_megamock=self,
                )
                # return_value is referenced when a call is made.
                # This actually accesses _mock_return_value
                if key == "_mock_return_value":
                    mega_result.megamock.name = f"{self.megamock.name}()"
                else:
                    mega_result.megamock.name = f"{self.megamock.name}.{key}"
                setattr(wrapped, key, mega_result)
                return mega_result
            return result
        if not self.megamock.spec and not self.megamock.wraps:
            return self._get_child_mock(_name=key)
        raise AttributeError(key)

    def __setattr__(self, key, value) -> None:
        if key in ("megamock", "_wrapped_legacy_mock"):
            self.__dict__[key] = value
            return
        if key in self.USE_SUPER_SETATTR:
            super().__setattr__(key, value)
        if self.megamock is not None:
            self._megamock_set_attr(key, value)
            self.megamock.attr_assignments[key].append(
                AttributeAssignment.for_current_stack(key, value)
            )
        else:
            self.__dict__[key] = value

    def _megamock_set_attr(self, key, value) -> None:
        if self.megamock.spy is not None:
            setattr(self.megamock.spy, key, value)
        elif (wrapped := self._wrapped_legacy_mock) is not None:
            try:
                setattr(wrapped, key, value)
            except AttributeError:
                # mock won't allow assignment of values assigned in __init__ or
                # elsewhere when spec_set is set. Check spec annotations and if the
                # value exists, allow assignment
                if key in self.megamock.spec.__annotations__:
                    self._set_attr_annotations_check(key, value)
                    wrapped.__dict__[key] = value
                else:
                    raise
        else:
            self.__dict__[key] = value

    def _set_attr_annotations_check(self, key: str, value: Any) -> None:
        # do not check type if assigning a mock object
        # note that MegaMock is a subclass of NonCallableMagicMock
        if not isinstance(value, mock.NonCallableMock | mock.NonCallableMagicMock):
            allowed_values = get_type_hints(self.megamock.spec)[key]

            def raise_type_error() -> None:
                raise TypeError(f"{value!r} is not an instance of {allowed_values}")

            if not isinstance(value, allowed_values):
                raise_type_error()

    def _get_spec_from_parents(
        self, _parent_stack: list[_MegaMockMixin] | None = None
    ) -> Any:
        """
        The built-in generate autospec function creates a few issues because
        it bypasses some of the tooling, so the whole tree of mocks are created
        without hitting MegaMock anywhere.

        This function walks up the megamock parent tree to find the root,
        then grabs the spec from there. It then walks back down, checking
        the children of the wrapped mock for each MegaMock to see if one
        of the legacy children mocks matches the wrapped mock from the
        MegaMock that called this function. It then grabs the attribute
        based on the name and the currently focused spec.
        """
        # to start, parent_stack is always empty
        parent_stack = _parent_stack or []

        # if this is a child mock, get the parent's spec
        if self.megamock.parent:
            cur_spec = self.megamock.parent._get_spec_from_parents(
                parent_stack + [self]
            )
            # abort if for some reason there is no parent spec
            if cur_spec is None:
                return None
        else:
            # the root MegaMock will hit this logic branch
            cur_spec = self.megamock.spec
        if cur_spec:
            if not (wrapped := self._wrapped_legacy_mock):
                return None
            if parent_stack:
                # the oldest item is the first element. The child that called this
                # will be the last element of parent_stack.
                child_that_called_me = parent_stack[-1]
                for child_name, obj in wrapped._mock_children.items():
                    if obj is child_that_called_me._wrapped_legacy_mock:
                        return getattr(cur_spec, child_name)
            else:
                return cur_spec
            # this is a class, and the child is an instance
            if isinstance(cur_spec, type):
                return cur_spec
        # this will happen if the root has no spec
        # or if the child that called this method cannot be located
        return None

    def __enter__(self) -> U:
        if self.megamock.wraps is not None:
            return self.megamock.wraps.__enter__()
        try:
            if self.using_real_logic():
                return self.megamock.spec.__enter__()  # type: ignore
            return self.__getattr__("__enter__")
        except AttributeError:
            if hasattr(self.megamock.spec, "__name__"):
                name_str = f"'{self.megamock.spec.__name__}' "  # type: ignore
            else:
                name_str = ""
            raise TypeError(
                f"{name_str}object does not support the context manager protocol"
            )

    def __exit__(self, exc_type, exc_value, traceback):
        if self.megamock.wraps is not None:
            self.megamock.wraps.__exit__(exc_type, exc_value, traceback)
        if self.using_real_logic():
            return self.megamock.spec.__exit__(exc_type, exc_value, traceback)
        return self.__getattr__("__exit__")

    def using_real_logic(self) -> bool:
        return (
            self._wrapped_legacy_mock.__dict__.get("_mock_return_value") is UseRealLogic
        )

    def _megalink_to(self, other: MegaMock[T, U]) -> None:
        """
        Link the call behavior to another mock. This is currently used by `MegaPatch.it`
        to link the class (type) mock with the class (instance) mock.
        """
        self.__dict__["_linked_mock"] = other

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} name='{self.megamock.name}' "
            f"| {self.meganame}>"
        )


class MegaMock(_MegaMockMixin[T, U], mock.MagicMock, Generic[T, U]):
    """
    The primary MegaMock class. Use MegaMock.it if providing a spec,
    spy, or wraps argument to ensure that type inference works correctly.

    Instead of: MegaMock(spec=MySpecClass)
    Use: MegaMock.it(MySpecClass)
    """

    # Types MegaMock()
    @overload
    def __init__(self: MegaMock[None, None]) -> None:
        ...

    # types MegaMock(spec=SomeClass, instance=False, ...)
    @overload
    def __init__(
        self: _MegaMockMixin[T, U],
        *,
        spec: type[T],
        instance: Literal[False] = False,
        side_effect: Any = None,
        return_value: Any = MISSING,
        wraps: None = None,
        spy: None = None,
        spec_set: bool = True,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ) = None,
        _parent_mega_mock: _MegaMockMixin | None = None,
        _merged_type: type[U],
    ) -> None:
        ...

    # types MegaMock(spec=SomeInstance)
    @overload
    def __init__(
        self: _MegaMockMixin[T, U],
        *,
        spec: T,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ) = None,
        instance: bool | None = None,
        side_effect: Any = None,
        return_value: Any = MISSING,
        wraps: Any = None,
        spy: Any = None,
        spec_set: bool = True,
        _parent_mega_mock: _MegaMockMixin | None = None,
        _merged_type: type[U],
    ) -> None:
        ...

    # types MegaMock when spec is not provided
    @overload
    def __init__(
        self: MegaMock[T, T],
        *,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ) = None,
        wraps: T | None = None,
        spy: T | None = None,
        spec_set: bool = True,
        instance: bool | None = None,
        side_effect: T | None = None,
        return_value: T | None = None,
        _parent_mega_mock: _MegaMockMixin | None = None,
        _merged_type: type[U] | None = None,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        spec: T | type[T] | None = None,
        wraps: T | None = None,
        spy: T | None = None,
        spec_set: bool = True,
        instance: bool | None = None,
        side_effect: Any = None,
        return_value: Any = MISSING,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ) = None,
        _parent_mega_mock: _MegaMockMixin | None = None,
        _merged_type: type[U] | None = None,
        # warning: kwargs to MagicMock may not work correctly! Use at your own risk!
        **kwargs,
    ) -> None:
        """
        :param spec: The MegaMock's specification (template object)
        :param wraps: An object to wrap, this is included for legacy support. Prefer spy
        :param spy: An object to spy on. The mock will maintain the real behavior of the
            object, but will also track attribute access and assignments
        :param spec_set: If True, only attributes in the spec will be allowed. Assigning
            attributes not part of the spec will result in a AttributeError
        :param instance: If True, the mock will be an instance of the spec. If False,
            the mock will be a class. By default this is True. This must be omitted or
            False for the AsyncMock
        :param side_effect: The side effect to use for the mock. Exceptoins are raised,
            fuctions are called, and iterables are returned in order in subsequent calls
        :param return_value: The return value to use for the mock.
        :param _wraps_mock: the wrapped mock, for internal use
        :param _parent_mega_mock: The parent MegaMock, for internal use
        """
        return super().__init__(
            spec,
            wraps=wraps,
            spy=spy,
            spec_set=spec_set,
            instance=instance,
            side_effect=side_effect,
            return_value=return_value,
            _wraps_mock=_wraps_mock,
            _parent_mega_mock=_parent_mega_mock,
            _merged_type=_merged_type,
            **kwargs,
        )

    @staticmethod
    def it(
        spec: T | None = None,
        *,
        wraps: T | None = None,
        spy: T | None = None,
        spec_set: bool = True,
        instance: bool | None = None,
        side_effect: Any = None,
        return_value: Any = MISSING,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ) = None,
        _parent_mega_mock: _MegaMockMixin | None = None,
        # warning: kwargs to MagicMock may not work correctly! Use at your own risk!
        **kwargs,
    ):
        # hack to get static type inference to think this is a true union
        # of the two classes
        def helper(obj) -> type[MegaMock[T, MegaMock | T] | T]:
            return cast(type[MegaMock | T], lambda: obj)

        return helper(
            MegaMock(
                spec=spec,
                wraps=wraps,
                spy=spy,
                spec_set=spec_set,
                instance=instance,
                side_effect=side_effect,
                return_value=return_value,
                _wraps_mock=_wraps_mock,
                _parent_mega_mock=_parent_mega_mock,
                _merged_type=type(MegaMock | spec.__class__),
                **kwargs,
            ),
        )()

    @staticmethod
    def from_legacy_mock(
        mock_obj: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
        ),
        spec: T,
        wraps: Any = None,
        parent_megamock: MegaMock | _MegaMockMixin | None = None,
    ):
        if not isinstance(mock_obj, (mock.MagicMock, mock.Mock)):
            if isinstance(mock_obj, mock.AsyncMock):
                return AsyncMegaMock(
                    _wraps_mock=mock_obj,
                    spec=spec,
                    wraps=wraps,
                    _parent_mega_mock=parent_megamock,
                )
            return NonCallableMegaMock.it(
                _wraps_mock=mock_obj,
                spec=spec,
                wraps=wraps,
                _parent_mega_mock=parent_megamock,
            )
        return MegaMock.it(
            _wraps_mock=mock_obj,
            spec=spec,
            wraps=wraps,
            _parent_mega_mock=parent_megamock,
        )

    def __call__(self, *args, **kwargs) -> Any:
        if wrapped := self._wrapped_legacy_mock:
            # cannot use wrapped.return_value because it will create a new
            # return_value object and use that
            # TODO: both the megamock and wrapped mock have return values
            #       and they're both importnat, we're not just looking at one.
            if self.using_real_logic():
                if not (spec := self.megamock.spec):
                    spec = self._get_spec_from_parents()
                    if not spec:
                        raise SpecRequiredException()
                if self.megamock.parent:
                    if hasattr(spec, "__func__"):
                        # convert from bound method to unbound method
                        return spec.__func__(self.megamock.parent, *args, **kwargs)
                    # instance of a class Mock
                    return cast(Callable, spec)(self.megamock.parent, *args, **kwargs)
                return cast(Callable, spec)(*args, **kwargs)
            if self._linked_mock is not None:
                # why is the mock not called as a bound method
                return self._linked_mock(self, *args, **kwargs)
            result = wrapped(*args, **kwargs)
            self._validate_spec_was_callable()
            if not isinstance(result, _MegaMockMixin) and isinstance(
                result, mock.NonCallableMock | mock.NonCallableMagicMock
            ):
                call_spec = self._get_call_spec()
                mega_result = MegaMock.from_legacy_mock(
                    result, call_spec, self.megamock.wraps, parent_megamock=self
                )
                return mega_result
            return result
        return super().__call__(*args, **kwargs)

    def _validate_spec_was_callable(self) -> None:
        """
        The create_autospec functionality properly creates a non-callable mock object.
        So, for example, MegaMock.it(Foo) properly indicates that the instance Foo is
        not callable.

        However, a nested call to a non-callable mock object will not be properly
        indicated as non-callable. For example, if we have a class Foo with a mocked
        method that returns a non-callable Bar, then the autospec'ed object will create
        a callable mock object in place of a mock Bar.

        In this case, _get_call_spec is used to give the mocked method a spec for a Bar
        instance, based on the return type annotation. This function then checks that
        spec and enforces the callability of it.

        In other words:

        MegaMock.it(Foo)()  <-- _validate_spec_was_callable is not needed
        MegaMock.it(Foo).do_something()() <-- _validate_spec_was_callable is needed
        """
        if self.megamock.wraps is not None:
            return
        if self.megamock.spec is None:
            return
        if not callable(self.megamock.spec):
            raise TypeError(f"{self.megamock.spec} is not callable")

    def _get_call_spec(self) -> Any:
        """
        Determine what the spec is for a function's return based on the annnotations.

        This is used to a create a mock object with the same traits as the spec.
        """
        if self.megamock.spec is None:
            return None
        if isclass(self.megamock.spec) and hasattr(self.megamock.spec, "__call__"):
            spec = self.megamock.spec.__call__
        else:
            spec = self.megamock.spec
        annotations = getattr(spec, "__annotations__", {})
        return_type = annotations.get("return", None)
        if return_type is None:
            return None
        return create_autospec(return_type, instance=True)


class NonCallableMegaMock(
    _MegaMockMixin[T, U], mock.NonCallableMagicMock, Generic[T, U]
):
    @staticmethod
    def it(
        spec: T | None = None,
        *,
        wraps: T | None = None,
        spy: T | None = None,
        spec_set: bool = True,
        instance: bool | None = None,
        side_effect: Any = None,
        return_value: Any = MISSING,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ),
        _parent_mega_mock: _MegaMockMixin | None = None,
        **kwargs,
    ):
        # hack to get static type inference to think this is a true union
        # of the two classes
        def helper(obj) -> type[NonCallableMegaMock[T, NonCallableMegaMock | T] | T]:
            return cast(type[NonCallableMegaMock | T], lambda: obj)

        return helper(
            NonCallableMegaMock(
                spec,
                wraps=wraps,
                spy=spy,
                spec_set=spec_set,
                instance=instance,
                side_effect=side_effect,
                return_value=return_value,
                _wraps_mock=_wraps_mock,
                _parent_mega_mock=_parent_mega_mock,
                _merged_type=type(NonCallableMegaMock | spec.__class__),
                **kwargs,
            ),
        )()


class AsyncMegaMock(MegaMock[T, U], mock.AsyncMock, Generic[T, U]):
    def __init__(self, *args, **kwargs) -> None:
        super(MegaMock, self).__init__(*args, **kwargs)

    def _get_child_mock(self, /, **kw) -> AsyncMegaMock:
        return AsyncMegaMock(**kw)

    @staticmethod
    def it(  # type: ignore
        spec: T | None = None,
        *,
        wraps: T | None = None,
        spy: T | None = None,
        spec_set: bool = True,
        instance: bool | None = None,
        side_effect: Any = None,
        return_value: Any = MISSING,
        _wraps_mock: (
            mock.Mock
            | mock.MagicMock
            | mock.NonCallableMock
            | mock.NonCallableMagicMock
            | None
        ),
        _parent_mega_mock: _MegaMockMixin | None = None,
        **kwargs,
    ):
        # hack to get static type inference to think this is a true union
        # of the two classes
        def helper(obj) -> type[AsyncMegaMock[T, AsyncMegaMock | T] | T]:
            return cast(type[AsyncMegaMock | T], lambda: obj)

        return helper(
            AsyncMegaMock(
                spec,
                wraps=wraps,
                spy=spy,
                spec_set=spec_set,
                instance=instance,
                side_effect=side_effect,
                return_value=return_value,
                _wraps_mock=_wraps_mock,
                _parent_mega_mock=_parent_mega_mock,
                _merged_type=type(AsyncMegaMock | spec.__class__),
                **kwargs,
            ),
        )



import builtins
import inspect
import linecache
import os
import re
import sys
import time
from types import FrameType, ModuleType
from typing import Callable

from megamock.import_references import References

MEASURE_TIMES = os.environ.get("MEASURE_TIMES", "0") == "1"

orig_import = builtins.__import__

skip_modules = {
    "typing",
    "functools",
    "asttokens",
}

perf_stats = {
    "num_imports": 0,
    "total_orig_import": 0.0,
    "total_if_check": 0.0,
    "total_get_frame": 0.0,
    "total_reconstruct": 0.0,
    "total_renamed_to": 0.0,
    "total_add_reference": 0.0,
}
perf_start_time = 0.0


if MEASURE_TIMES:

    def measure_start() -> None:
        global perf_start_time
        perf_start_time = time.time()

    def measure(name: str) -> None:
        global perf_start_time
        perf_stats[name] += time.time() - perf_start_time

else:

    def measure_start() -> None:
        pass

    def measure(name: str) -> None:
        pass


pat = re.compile(r"^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))|^(\s*@)")


def findsource(object):
    """
    A streamlined version of inspect.findsource
    """

    file = inspect.getsourcefile(object)
    if file:
        # Invalidate cache if needed.
        linecache.checkcache(file)
    else:
        file = inspect.getfile(object)
        # Allow filenames in form of "<something>" to pass through.
        # `doctest` monkeypatches `linecache` module to enable
        # inspection, so let `linecache.getlines` to be called.
        if not (file.startswith("<") and file.endswith(">")):
            raise OSError("source code not available")

    module = inspect.getmodule(object, file)
    if module:
        lines = linecache.getlines(file, module.__dict__)
    else:
        lines = linecache.getlines(file)
    if not lines:
        raise OSError("could not get source code")
    return lines


def _get_code_lines(frame: FrameType) -> tuple[list[str], int]:
    """
    Streamlined logic to get code lines from a frame
    """
    lines = findsource(frame)
    lineno = frame.f_lineno
    start = lineno - 1
    start = max(0, min(start, len(lines) - 1))
    code_lines = lines[start : start + 1]

    return code_lines, lineno


def _reconstruct_full_line(
    frame: FrameType, getline: Callable = linecache.getline
) -> str:
    """
    Given an import frame, reconstruct the full import line, which
    may span multiple lines of code via parenthesis or backslashes

    :param frame: The frame of the import
    :param getline: Pytest uses linecache.getline to rewrite assertions, so providing
        getline as a parameter for testing purposes
    """

    code_lines, lineno = _get_code_lines(frame)
    if code_lines:
        # if this is a multiline import, reconstruct the full line
        # the absense of '(' or '\' indicates a single line import
        next_line = code_lines[0]
        # note: timit gives 0.078569 for this vs 0.134779 for a compiled regex
        if "(" in next_line or "\\" in next_line:
            filename = frame.f_code.co_filename
            linenum = lineno + 1

            lines = [next_line]
            paren_count = next_line.count("(") - next_line.count(")")
            while paren_count > 0 or next_line.rstrip("\n\r").endswith("\\"):
                next_line = getline(filename, linenum)
                lines.append(next_line)
                paren_count -= next_line.count(")")
                linenum += 1
            return "".join(lines)
        return next_line
    return ""


has_rename_regex = re.compile(r"\s*from \S+ import.*\s+as\s", re.DOTALL)


def start_import_mod() -> None:
    """
    Start the import modification

    This should be done as one of the first things when testing
    """

    def new_import(*args, **kwargs) -> ModuleType:
        perf_stats["num_imports"] += 1
        measure_start()
        result = orig_import(*args, **kwargs)
        measure("total_orig_import")

        measure_start()
        module_name = args[0]
        proceed = (
            module_name not in skip_modules
            and not module_name.startswith("_")
            and (target_module := sys.modules.get(module_name))
            and len(args) > 3
            and (names := args[3])
        )
        measure("total_if_check")
        if proceed:
            assert target_module is not None
            measure_start()
            for i in range(1, 5):
                frame = sys._getframe(i)
                if frame.f_code.co_name == "new_import":
                    continue
                calling_module = inspect.getmodule(frame)
                if calling_module:
                    break
            measure("total_get_frame")
            assert calling_module
            measure_start()
            full_line = _reconstruct_full_line(frame)
            measure("total_reconstruct")
            for k in names:
                measure_start()
                if full_line and (
                    has_rename_regex.search(full_line)
                    and (
                        renamed_result := re.search(
                            rf"\s*from \S+ import.*{k}\s+as\s+(\w+)",
                            full_line,
                            re.DOTALL,
                        )
                    )
                ):
                    renamed_to = renamed_result.group(1)
                else:
                    renamed_to = k
                measure("total_renamed_to")
                measure_start()
                References.add_reference(target_module, calling_module, k, renamed_to)
                measure("total_add_reference")

        return result

    builtins.__import__ = new_import



from collections import defaultdict
from types import ModuleType
from typing import NamedTuple

ModAndName = NamedTuple("ModAndName", [("module", str), ("name", str)])


class References:
    """
    The References class is used as part of the import machinary and its
    the magic that allows MegaPatch to work by simply passing things in.
    """

    # References from a calling module and the name used to the
    # source module and the original name
    references: dict[str, dict[str, ModAndName]] = defaultdict(dict)
    # reverse references from the source module and the original,
    # non-nested name, to the calling module and the name used
    reverse_references: dict[str, dict[str, set[ModAndName]]] = defaultdict(
        lambda: defaultdict(set)
    )
    # renames from the calling module and the name used to the original name
    renames: dict[ModAndName, str] = {}

    @staticmethod
    def add_reference(
        module: ModuleType,
        calling_module: ModuleType,
        original_name: str,
        named_as: str,
    ) -> None:
        # do not bother with bad modules. Example: conftest from pytest
        if not calling_module.__package__:
            # this can happen if using megamock + pytest hot reloader
            return
        References._add_reference(module, calling_module, original_name, named_as)

    @staticmethod
    def _add_reference(
        module: ModuleType,
        calling_module: ModuleType,
        original_name: str,
        named_as: str,
    ) -> None:
        module_path = module.__name__
        References.references[calling_module.__name__][named_as] = ModAndName(
            module_path,
            original_name,
        )
        base_original_name = original_name.split(".")[0]
        References.reverse_references[module_path][base_original_name].add(
            ModAndName(calling_module.__name__, named_as)
        )
        if original_name != named_as:
            References.renames[
                ModAndName(calling_module.__name__, named_as)
            ] = original_name

    @staticmethod
    def get_references(module_name: str, named_as: str) -> set[ModAndName]:
        """
        Given an importing module and name used, return the source module and name

        A set is used, but it can't have more than one element
        """
        val = References.references[module_name].get(named_as)
        if not val:
            return set()
        return {val}

    @staticmethod
    def get_reverse_references(module_name: str, original_name: str) -> set[ModAndName]:
        """
        Given a source module and original name, return the imports to it.
        This includes logic to handle the nesting of the original name. For example,
        Foo.bar may be referenced as OtherFoo.bar. Since "Foo.bar" isn't kept track of
        for imports, only "Foo", the base name is extracted from original_name and then
        the referenced name is rebuilt from the right hand side of original_name.

        So for example, the reverse_references may look like this:
            {
                "example.module": {
                    "Foo": {
                        ModAndName("example.other.module", "OtherFoo")
                    }
                }
            }

        get_reverse_references("example.module", "Foo.bar") would return:
            {
                ("example.other.module", "OtherFoo.bar")
            }

        The import that got us here would look like this:
            from example.other.module import Foo as OtherFoo

        And then the MegaPatch might look like this:
            MegaPatch.it(OtherFoo.bar, ...)
        """
        components = original_name.split(".", 0)
        if len(components) > 1:
            base_name, right_side = components[0], [".".join(components[1:])]
        else:
            base_name = components[0]
            right_side = []
        return {
            ModAndName(x.module, ".".join([x.name] + right_side))
            for x in References.reverse_references[module_name].get(base_name, set())
        }

    @staticmethod
    def get_original_name(module_name: str, named_as: str) -> str:
        """
        Given an importing module and name used, return the original name
        """
        return References.renames.get(ModAndName(module_name, named_as), named_as)



from typing import Callable, Sequence, cast

from megamock.megamocks import MegaMock, UseRealLogic
from megamock.type_util import Call


class Mega:
    """
    Wrapper class around callable MegaMock objects. This provides
    a convenience to the developer via autocomplete, a more
    standard interface, and tools for values against cast'ed objects.

    Instead of (does not autocomplete "assert_called_with"):
        mock.my_method.assert_called_with(1, 2, 3)

    Use (autocompletes "assert_called_with"):
        assert Mega(mock.my_method).called_with(1, 2, 3)
    """

    last_assertion_error: AssertionError | None = None

    def __init__(self, func: Callable) -> None:
        self._func = func

    def _check_mock_assertion(self, func: Callable) -> bool:
        try:
            func()
        except AssertionError as err:
            Mega.last_assertion_error = err
            return False
        return True

    @property
    def _mm(self) -> MegaMock:
        """
        Property to shorthand the cast to MegaMock
        """
        return cast(MegaMock, self._func)

    def called_once_with(self, *args, **kwargs) -> bool:
        """
        Return true if the mock was called exactly once and with the specified
        arguments
        """
        return self._check_mock_assertion(
            lambda: self._mm.assert_called_once_with(*args, **kwargs)
        )

    def called_once(self) -> bool:
        """
        Return true if the mock was called exactly once
        """
        return self._check_mock_assertion(lambda: self._mm.assert_called_once())

    def called(self) -> bool:
        """
        Return true if the mock was called
        """
        return self._mm.called

    def not_called(self) -> bool:
        """
        Return true if the mock was not called
        """
        return self._check_mock_assertion(lambda: self._mm.assert_not_called())

    def called_with(self, *args, **kwargs) -> bool:
        """
        Return true if the last call made to the mock was with the specified
        arguments
        """
        return self._check_mock_assertion(
            lambda: self._mm.assert_called_with(*args, **kwargs)
        )

    def any_call(self, *args, **kwargs) -> bool:
        """
        Return true if the mock was called with the specified arguments
        at any point in time
        """
        return self._check_mock_assertion(
            lambda: self._mm.assert_any_call(*args, **kwargs)
        )

    def has_calls(self, calls: Sequence[Call], any_order=False) -> bool:
        """
        Return true if all of the given calls were made.

        If any_order is true, the calls can be made in any order.

        Extra calls are ignored.
        """
        try:
            iter(calls)
        except TypeError:
            raise TypeError("First argument must be an iterable of call objects")
        return self._check_mock_assertion(
            lambda: self._mm.assert_has_calls(calls, any_order)
        )

    @property
    def call_args(self) -> Call:
        """
        Return the last call made to the mock
        """
        return self._mm.call_args

    @property
    def call_args_list(self) -> list[Call]:
        """
        Return the list of all calls made to the mock
        """
        return self._mm.call_args_list

    @property
    def call_count(self) -> int:
        """
        Return the number of times the mock was called
        """
        return self._mm.call_count

    def use_real_logic(self) -> None:
        """
        Helper function to have a MegaMock object use the real logic
        """
        self._mm.return_value = UseRealLogic
